@@ -25,6 +25,7 @@ import (
 	"golang.org/x/tools/go/analysis/passes/deepequalerrors"
 	"golang.org/x/tools/go/analysis/passes/errorsas"
 	"golang.org/x/tools/go/analysis/passes/httpresponse"
+	"golang.org/x/tools/go/analysis/passes/ifaceassert"
 	"golang.org/x/tools/go/analysis/passes/loopclosure"
 	"golang.org/x/tools/go/analysis/passes/lostcancel"
 	"golang.org/x/tools/go/analysis/passes/nilfunc"
@@ -32,6 +33,7 @@ import (
 	"golang.org/x/tools/go/analysis/passes/shift"
 	"golang.org/x/tools/go/analysis/passes/sortslice"
 	"golang.org/x/tools/go/analysis/passes/stdmethods"
+	"golang.org/x/tools/go/analysis/passes/stringintconv"
 	"golang.org/x/tools/go/analysis/passes/structtag"
 	"golang.org/x/tools/go/analysis/passes/testinggoroutine"
 	"golang.org/x/tools/go/analysis/passes/tests"
@@ -114,14 +116,15 @@ func DefaultOptions() *Options {
 					CommandUpgradeDependency.Name: true,
 					CommandVendor.Name:            true,
 				},
-				LinksInHover:            true,
-				CompleteUnimported:      true,
-				CompletionDocumentation: true,
-				DeepCompletion:          true,
-				ImportShortcut:          Both,
-				Matcher:                 Fuzzy,
-				SymbolMatcher:           SymbolFuzzy,
-				SymbolStyle:             PackageQualifiedSymbols,
+				LinksInHover:                true,
+				CompleteUnimported:          true,
+				CompletionDocumentation:     true,
+				DeepCompletion:              true,
+				ImportShortcut:              Both,
+				Matcher:                     Fuzzy,
+				SymbolMatcher:               SymbolFuzzy,
+				SymbolStyle:                 PackageQualifiedSymbols,
+				ExperimentalPackageCacheKey: true,
 			},
 			InternalOptions: InternalOptions{
 				LiteralCompletions: true,
@@ -162,6 +165,8 @@ type ClientOptions struct {
 	PreferredContentFormat            protocol.MarkupKind
 	LineFoldingOnly                   bool
 	HierarchicalDocumentSymbolSupport bool
+	SemanticTypes                     []string
+	SemanticMods                      []string
 }
 
 // ServerOptions holds LSP-specific configuration that is provided by the
@@ -350,9 +355,12 @@ type ExperimentalOptions struct {
 	// semantic tokens to the client.
 	SemanticTokens bool
 
-	// ExpandWorkspaceToModule instructs `gopls` to expand the scope of the workspace to include the
-	// modules containing the workspace folders. Set this to false to avoid loading
-	// your entire module. This is particularly useful for those working in a monorepo.
+	// ExpandWorkspaceToModule instructs `gopls` to adjust the scope of the
+	// workspace to find the best available module root. `gopls` first looks for
+	// a go.mod file in any parent directory of the workspace folder, expanding
+	// the scope to that directory if it exists. If no viable parent directory is
+	// found, gopls will check if there is exactly one child directory containing
+	// a go.mod file, narrowing the scope to that directory if it exists.
 	ExpandWorkspaceToModule bool
 
 	// ExperimentalWorkspaceModule opts a user into the experimental support
@@ -535,6 +543,12 @@ func (o *Options) ForClientCapabilities(caps protocol.ClientCapabilities) {
 	o.LineFoldingOnly = fr.LineFoldingOnly
 	// Check if the client supports hierarchical document symbols.
 	o.HierarchicalDocumentSymbolSupport = caps.TextDocument.DocumentSymbol.HierarchicalDocumentSymbolSupport
+	// Check if the client supports semantic tokens
+	o.SemanticTypes = caps.TextDocument.SemanticTokens.TokenTypes
+	o.SemanticMods = caps.TextDocument.SemanticTokens.TokenModifiers
+	// we don't need Requests, as we support full functionality
+	// we don't need Formats, as there is only one, for now
+
 }
 
 func (o *Options) Clone() *Options {
@@ -597,7 +611,6 @@ func (o *Options) AddStaticcheckAnalyzer(a *analysis.Analyzer) {
 // enableAllExperimentMaps.
 func (o *Options) enableAllExperiments() {
 	o.ExperimentalDiagnosticsDelay = 200 * time.Millisecond
-	o.ExperimentalPackageCacheKey = true
 	o.SymbolStyle = DynamicSymbols
 }
 
@@ -951,31 +964,33 @@ func convenienceAnalyzers() map[string]Analyzer {
 func defaultAnalyzers() map[string]Analyzer {
 	return map[string]Analyzer{
 		// The traditional vet suite:
-		asmdecl.Analyzer.Name:      {Analyzer: asmdecl.Analyzer, Enabled: true},
-		assign.Analyzer.Name:       {Analyzer: assign.Analyzer, Enabled: true},
-		atomic.Analyzer.Name:       {Analyzer: atomic.Analyzer, Enabled: true},
-		atomicalign.Analyzer.Name:  {Analyzer: atomicalign.Analyzer, Enabled: true},
-		bools.Analyzer.Name:        {Analyzer: bools.Analyzer, Enabled: true},
-		buildtag.Analyzer.Name:     {Analyzer: buildtag.Analyzer, Enabled: true},
-		cgocall.Analyzer.Name:      {Analyzer: cgocall.Analyzer, Enabled: true},
-		composite.Analyzer.Name:    {Analyzer: composite.Analyzer, Enabled: true},
-		copylock.Analyzer.Name:     {Analyzer: copylock.Analyzer, Enabled: true},
-		errorsas.Analyzer.Name:     {Analyzer: errorsas.Analyzer, Enabled: true},
-		httpresponse.Analyzer.Name: {Analyzer: httpresponse.Analyzer, Enabled: true},
-		loopclosure.Analyzer.Name:  {Analyzer: loopclosure.Analyzer, Enabled: true},
-		lostcancel.Analyzer.Name:   {Analyzer: lostcancel.Analyzer, Enabled: true},
-		nilfunc.Analyzer.Name:      {Analyzer: nilfunc.Analyzer, Enabled: true},
-		printf.Analyzer.Name:       {Analyzer: printf.Analyzer, Enabled: true},
-		shift.Analyzer.Name:        {Analyzer: shift.Analyzer, Enabled: true},
-		stdmethods.Analyzer.Name:   {Analyzer: stdmethods.Analyzer, Enabled: true},
-		structtag.Analyzer.Name:    {Analyzer: structtag.Analyzer, Enabled: true},
-		tests.Analyzer.Name:        {Analyzer: tests.Analyzer, Enabled: true},
-		unmarshal.Analyzer.Name:    {Analyzer: unmarshal.Analyzer, Enabled: true},
-		unreachable.Analyzer.Name:  {Analyzer: unreachable.Analyzer, Enabled: true},
-		unsafeptr.Analyzer.Name:    {Analyzer: unsafeptr.Analyzer, Enabled: true},
-		unusedresult.Analyzer.Name: {Analyzer: unusedresult.Analyzer, Enabled: true},
+		asmdecl.Analyzer.Name:       {Analyzer: asmdecl.Analyzer, Enabled: true},
+		assign.Analyzer.Name:        {Analyzer: assign.Analyzer, Enabled: true},
+		atomic.Analyzer.Name:        {Analyzer: atomic.Analyzer, Enabled: true},
+		bools.Analyzer.Name:         {Analyzer: bools.Analyzer, Enabled: true},
+		buildtag.Analyzer.Name:      {Analyzer: buildtag.Analyzer, Enabled: true},
+		cgocall.Analyzer.Name:       {Analyzer: cgocall.Analyzer, Enabled: true},
+		composite.Analyzer.Name:     {Analyzer: composite.Analyzer, Enabled: true},
+		copylock.Analyzer.Name:      {Analyzer: copylock.Analyzer, Enabled: true},
+		errorsas.Analyzer.Name:      {Analyzer: errorsas.Analyzer, Enabled: true},
+		httpresponse.Analyzer.Name:  {Analyzer: httpresponse.Analyzer, Enabled: true},
+		ifaceassert.Analyzer.Name:   {Analyzer: ifaceassert.Analyzer, Enabled: true},
+		loopclosure.Analyzer.Name:   {Analyzer: loopclosure.Analyzer, Enabled: true},
+		lostcancel.Analyzer.Name:    {Analyzer: lostcancel.Analyzer, Enabled: true},
+		nilfunc.Analyzer.Name:       {Analyzer: nilfunc.Analyzer, Enabled: true},
+		printf.Analyzer.Name:        {Analyzer: printf.Analyzer, Enabled: true},
+		shift.Analyzer.Name:         {Analyzer: shift.Analyzer, Enabled: true},
+		stdmethods.Analyzer.Name:    {Analyzer: stdmethods.Analyzer, Enabled: true},
+		stringintconv.Analyzer.Name: {Analyzer: stringintconv.Analyzer, Enabled: true},
+		structtag.Analyzer.Name:     {Analyzer: structtag.Analyzer, Enabled: true},
+		tests.Analyzer.Name:         {Analyzer: tests.Analyzer, Enabled: true},
+		unmarshal.Analyzer.Name:     {Analyzer: unmarshal.Analyzer, Enabled: true},
+		unreachable.Analyzer.Name:   {Analyzer: unreachable.Analyzer, Enabled: true},
+		unsafeptr.Analyzer.Name:     {Analyzer: unsafeptr.Analyzer, Enabled: true},
+		unusedresult.Analyzer.Name:  {Analyzer: unusedresult.Analyzer, Enabled: true},
 
 		// Non-vet analyzers:
+		atomicalign.Analyzer.Name:      {Analyzer: atomicalign.Analyzer, Enabled: true},
 		deepequalerrors.Analyzer.Name:  {Analyzer: deepequalerrors.Analyzer, Enabled: true},
 		sortslice.Analyzer.Name:        {Analyzer: sortslice.Analyzer, Enabled: true},
 		testinggoroutine.Analyzer.Name: {Analyzer: testinggoroutine.Analyzer, Enabled: true},
